/* cc -O2 -o kostnice kostnice.c
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
128..191
8901234567890123456789012345678901234567890123456789012345678901


192..255
2345678901234567890123456789012345678901234567890123456789012345

*/

void help(int verb)
{
  if (verb) 
    fprintf(stderr,"\
Konverze kodu pro cestinu.\n\
Pouziti:\n\
  kostnice [VSTUPNI_SOUBOR] [VSTUPNI_KOD] [VYSTUPNI_SOUBOR] [VYSTUPNI_KOD]\n\
nebo\n\
  kostnice [VSTUPNI_KOD] [VSTUPNI_SOUBOR] [VYSTUPNI_KOD] [VYSTUPNI_SOUBOR]\n\
KOD:\n\
  -a ASCII 7bit        vstupni: klavesnice ``` (carka=`, hacek=``, umlaut=```)\n\
                       vystupni: bez diakritiky\n\
  -c Cork              (te vak zm esky)\n\
  -d DKOI 8            (te vak zm esky)\n\
  -e echo ve 2 ASCII radcich na stderr, lze kombinovat\n\
  -i Unix ISO 8859-2   (te vak zm esky)\n\
  -k Kamenicky         (te vak zm esky)\n\
  -l PC Latin II CP852 (te vak zm esky), DOS, OS/2, Novell\n\
  -n potlaci (CR)LF na konci radku\n\
  -p = -l\n\
  -u = -i\n\
  -w Windows CP1250    (te vak zm esky)\n");
  else
    fprintf(stderr,"\
Konverze kodu pro cestinu. Pouziti:\n\
  kostnice [VSTUPNI_SOUBOR] [VSTUPNI_KOD] [VYSTUPNI_SOUBOR] [VYSTUPNI_KOD]\n\
KOD:\n\
  -a ASCII 7bit        vstupni: klavesnice ``` (carka=`, hacek=``, umlaut=```)\n\
                       vystupni: bez diakritiky\n\
  -c Cork\n\
  -d DKOI 8\n\
  -e echo ve 2 ASCII radcich na stderr, lze kombinovat\n\
  -h help (vc. vzorku textu)\n\
  -i Unix ISO 8859-2\n\
  -k Kamenicky\n\
  -l PC Latin II CP852 (DOS, OS/2, Novell)\n\
  -n potlaci (CR)LF na konci radku\n\
  -w Windows CP1250\n");

  fprintf(stderr,"\
Pouze - jako soubor znaci stdin nebo stdout\n\
Defaulty:\n\
  VSTUPNI_SOUBOR=stdin\n"
#ifdef DOS
"  VSTUPNI_KOD=-k\n"
#else
"  VSTUPNI_KOD=-u\n"
#endif
"  VYSTUPNI_SOUBOR=stdout\n\
  VYSTUPNI_KOD=-a\n\
");
exit(0); 
}

FILE *in,*out;
static char *echo=NULL,*mod=NULL;

void ascecho(char c,char m)
{
static int pos;

if (echo==NULL || mod==NULL) return;
echo[pos]=c; mod[pos]=m;
if (c=='\n') mod[pos]='\n';
pos++;
if (pos>=1024) {
  fprintf(stderr,"prilis dlouha radka\n");
  exit(-1); }
echo[pos]=mod[pos]=0;
if (c=='\n') {
  fputs(mod,out);
  fputs(echo,out);
  pos=0; }
}


int main(int narg,char **arg)
{
int c,ch,i,m,x;
char *infn=NULL,*outfn=NULL;
int incode=-1,outcode='a',kbd=0,NL=1;

/* metacode:
   Xc[.] je znak bez diakritiky
   Xm[.] je diakritika:
     , = carka
     : = prehlaska
     H = prehlaska (sikma, neexistuje-li, prevadena na :)
     v = hacek
     o = krouzek
     u = obloucek
     / = skrtnuto
     > = cedilla apod.
     ^ = striska
     . = tecka
     - = pruh
     ` = acute (carka zpet)
     ~ = vlnovka
     x = times (neexistuje-li, preveden na x)
     ! = paragraf (znak = S)
   Upozorneni: garantovano je pouze prevadeni ceskych a slovenskych znaku!
*/
char kc[256],km[256]; /* Kamenicky */
char lc[256],lm[256]; /* PC Latin II */
char dc[256],dm[256]; /* KOI 8 */
char wc[256],wm[256]; /* Windows 1250 */
char uc[256],um[256]; /* Unix ISO 8859-2 */
char cc[256],cm[256]; /* Cork */
char *ic,*im; /* input metacode */
char *oc,*om; /* output metacode */


/* Kamenicky */
for (c=0; c<256; c++) kc[c]=c; memset(km,32,256);
memcpy(kc+128,"CuedaDTceELIllAAEzZooOuUyOUSLYRtaiounNUOsrrRS",46);
/*              */
memcpy(km+128,"v:,v:vvvvv,,v,:,,vv^:,o,,::vv,vv,,,,vvo^vv,, !",46);


/* PC Latin II */
for (c=0; c<256; c++) lc[c]=c; memset(lm,32,256);
/*              */
memcpy(lc+128,"CueaauccleOoiZACELlooLlSsOUTtL caiouAaZzEe zCsAAESZz",64);
memcpy(lm+128,">:,^:o,>/:::^,:,,,,^:vv,,::vv/xv,,,,>>vv>> ,v>\"\"       ,^v>    .. ",64);

/*             23456789012345678901234567890123456789012345678901   23456789   012345 */
/*              */
memcpy(lc+192,"Aa dDDEdNIIeTUOONnnSsRUrUyYt   " "   S" "uRr",64);
memcpy(lm+192,"      uu       $--v:vv,^v    >o , ^,,vvv,,,:,,>,-\""",vu! >o\""":vv  ",64);


/* KOI 8 */
for (c=0; c<256; c++) dc[c]=c; memset(dm,32,256);
/*              */
memcpy(dc+128,"CueSᜒAa  $oiNn       S",64);
memcpy(dm+128,">``!    o  o   ``  ~ ~                       !                  ",64);

/*             2345678901234567890123456789012345678901234567890123456789012345 */
memcpy(dc+192,"acderuiullonooarstueayzv,oſACDERUIULLONOOARSTUEAYZ\"^ ",64);
memcpy(dm+192," , vvv, :,o,v:v,^:vvv, ,`,v      , vvv, :,o,v:v,^:vvv, ,`,v    ",64);


/* Windows 1250 */
for (c=0; c<256; c++) wc[c]=c; memset(wm,32,256);
/*                 =uvozovky dole */
/*              */
memcpy(wc+128,"  ' \"     S<STZZ ''\"\""
                                   " -  s>stzz   L A|S  S    Z   l     as L lz",64);
memcpy(wm+128,"          v ,vv,       -  v ,vv, vu/$> !\" >\""
                                                          " - .o >/,   >>>\"v v.",64);

/*              */
memcpy(wc+192,"RAAAALCCCEEEEIIDDNNOOOOxRUUUUYTSraaaalccceeeeiiddnnoooo ruuuuyt",64);
memcpy(wm+192,",,^v:,,>v,>:v,^v/,v,^H: vo,H:,>!,,^v:,,>v,>:v,^v/,v,^H: vo,H:,>",64);


/* ISO 8859-2 (UNIX) */
for (c=0; c<256; c++) uc[c]=c; memset(um,32,256);
/*             8901234567890123456789012345678901234567890123456789012345678901 */
memcpy(uc+128,"                  '              A L$LS  SSTZ ZZ a l ls  sstz zz",64);
memcpy(um+128,"                                 >v/ v,!:v>v, v.o>>/,v,v>v>v,Hv.",64);

/*             2345678901234567890123456789012345678901234567890123456789012345 */
/*              */
memcpy(uc+192,"RAAAALCCCEEEEIIDDNNOOOOxRUUUUYTSraaaalccceeeeiiddnnoooo:ruuuuyt",64);
memcpy(um+192,",,^v:,,>v,>:v,^v/,v,^H: vo,H:,>!,,^v:,,>v,>:v,^v/,v,^H: vo,H:,>",64);


/* Cork */
for (c=0; c<256; c++) cc[c]=c; memset(cm,32,256);
/*             8901234567890123456789012345678901234567890123456789012345678901 */
memcpy(cc+128,"AACCDEEGLLLNN ORRSSSTTUUYZZZ IdSaaccdeeglllnnnorrsssttuuyzzz !?$",64);
memcpy(cm+128,"u>,vvv>u'v/,v :,v,v>v>:o:,v. ./!u>,vvv>u,v/,v>:,v,v>v>:o:,v.    ",64);

/*             2345678901234567890123456789012345678901234567890123456789012345 */
memcpy(cc+192,"AAAAAAACEEEEIIIIDNOOOOOOOUUUUY Saaaaaaaceeeeiiiidnooooooouuuuy s",64);
memcpy(cm+192,"`,^~:o >`,^:`,^:/~`,^~: /`,^:,  `,^~:o >`,^:`,^:/~`,^~: /`,^:,  ",64);

if (narg<2) help(0);

for (i=1; i<narg; i++)
  if (arg[i][0]=='-' && arg[i][1])
    if (arg[i][1]=='h') help(1);
    else if (arg[i][1]=='e') {
      echo=malloc(1024); mod=malloc(1024); }
    else if (arg[i][1]=='n') NL=0;
    else  
      if (incode==-1) incode=arg[i][1];
      else outcode=arg[i][1];
  else
    if (infn==NULL) infn=arg[i];
    else outfn=arg[i];

if (incode==-1) {
#ifdef DOS
  incode='k';
#else
  incode='u';
#endif
  fprintf(stderr,"default vstupni kod = -%c\n",incode);
  }

if (!infn || !strcmp(infn,"-")) in=stdin;
else if (!(in=fopen(infn,"rt"))) {
  fprintf(stderr,"vstupni soubor nelze otevrit\n");
  exit(-1); }

if (!outfn || !strcmp(outfn,"-")) out=stdout;
else if (!strcmp(infn,outfn)) {
  fprintf(stderr,"vstupni soubor == vystupni soubor\n");
  exit(-1); }
else if (!(out=fopen(outfn,"wt"))) {
  fprintf(stderr,"vystupni soubor nelze otevrit\n");
  exit(-1); }

switch (incode) {
  case 'a': kbd++; break;
  case 'k': im=km; ic=kc; break;
  case 'p':
  case 'l': im=lm; ic=lc; break;
  case 'd': im=dm; ic=dc; break;
  case 'w': im=wm; ic=wc; break;
  case 'i':
  case 'u': im=um; ic=uc; break;
  case 'c': im=cm; ic=cc; break;
  default: help(0); }

switch (outcode) {
  case 'k': om=km; oc=kc;
  case 'a': break;
  case 'p':
  case 'l': om=lm; oc=lc; break;
  case 'd': om=dm; oc=dc; break;
  case 'w': om=wm; oc=wc; break;
  case 'i':
  case 'u': om=um; oc=uc; break;
  case 'c': om=cm; oc=cc; break;
  default: help(0); }

if (ic==oc && !echo) {
  fprintf(stderr,"vstupni kod = vystupni kod\n");
  exit(-1); }

if (outcode=='a' && !echo) while ( (c=fgetc(in))>=0 ) {
  c=ic[c];
  fputc(c,out);
  ascecho(c,m); }

else while ( (ch=fgetc(in))>=0 ) {

  /* patch to replace ^L by ^M^J^M */
  if (outcode=='w') if (ch==12) ch=10;

  if (kbd) {
    /* ``` keyboard */
    switch (ch) {
      case '0': m=','; x='e'; break;
      case '1': m=' '; x=';'; break;
      case '2': m='v'; x='e'; break;
      case '3': m='v'; x='s'; break;
      case '4': m='v'; x='c'; break;
      case '5': m='v'; x='r'; break;
      case '6': m='v'; x='z'; break;
      case '7': m=','; x='y'; break;
      case '8': m=','; x='a'; break;
      case '9': m=','; x='i'; break;
      case ';': m='o'; x='u'; break; 
      case '`': m=',';
	ch=fgetc(in);
        if (strchr("dtnDTN",ch)) m='v';
	if (ch=='`') {
	  ch=fgetc(in);
	  m='v';
	  if (ch=='u' || ch=='U') m='o';
	  if (ch=='o' || ch=='O') m='^';
	  if (ch=='`') {
	    m=':';
	    ch=fgetc(in);
	    if (ch=='`') {
	      x='`'; goto found; }
	    else x=ch; }
	  else x=ch; }
	else x=ch;
	break;
      default: m=' '; x=ch; break;
      }
    c=x; }
  else {
    m=im[ch]; x=c=ic[ch]; }

  if (outcode>0 && outcode!='a') {
    for (i=0; i<256; i++) if (oc[i]==c && om[i]==m) { x=i; goto found; }
    if (m=='x') { x='x'; goto found; } /* times */
    if (m=='H') {
      /* nenasla se sikma prehlaska, hledame kolmou */
      m=':';
      for (i=0; i<256; i++) if (oc[i]==c && om[i]==m) { x=i; goto found; } }
    if (m==' ') x=' '; /* character does not exist */
    /* patch for line draw characters translated to | + etc. in win
       note that  is long - */
    if (outcode=='w') {
      static char tr[]="\0"
	"||||  ||      |+||  |+-    ++  ";
      char *trc;
      trc=strchr(tr,ch);
      if (trc) x=trc[41]; }
   found: 
    if (x!='\n' || NL) fputc(x,out); 
    }
  ascecho(c,m); }
  
if (outcode=='e') ascecho('\n',' ');
fclose(out);
fclose(in);

return 0;
}
