/* cc -O2 -o l2h l2h.c

  08/2015:
    - djpeg replaced by jpegtopnm
    - some html entities added

  11/2008:
    - audio version
  10/2008: 
    - gamma for ppmscale set to 1 (default 2.2 gives too light text)
    - DPI increased from 96 to 108
  
 l2h = 1st pass of LaTeX->HTML translation
 - creates files like p.1.1.1.html
 see also:
   makeall.sh
   l2h2 (2nd pass - called directly from here)
   l2h3 (3rd pass)
   
 (c) J. Kolafa 2002-5,2008,2015 jiri.kolafa@vscht.cz
 NOTE: I had to write this program because the LaTeX2HTML translator
       failed to convert more complex and longer LaTeX files...

 BUG: schem, sunit have wrong size in generated GIFs

 new CSS-based version */

#include "/home/jiri/macsimus/gen/include.h"
typedef unsigned unsigned4;
#include "/home/jiri/macsimus/gen/crcgen.c"
#include "/home/jiri/macsimus/gen/mystrdup.c"

#define LINELEN 1024

#define MAXNEST 32 /* of {} \begin{}..\end or so */
int DPI=108,SIZE;
int SCALE=3; /* antialiasing for generating GIFs */
FILE *soundconfig; /* .xml, also switch if sound */
char *ppmtogif="ppmtogif -transparent \"#ffffff\"";
char *colorppmtogif="ppmquant 256 | ppmtogif -transparent \"#ffffff\""; // -C
#define CELLWIDTH "742" /* of middle cell of the table: must be in "" */
#define TABWIDTH "760" /* CELLWIDTH+2*9 (9=width of corners) */
#define LINEWIDTH "740" /* of line dividing the table: must be in ""; also for TeX->GIF */
#define MAXITEMS 1000000 /* for cycle checks... */
#define EQSEP "&nbsp;" /* separator of = < > in math */
#define MENUSEP " &nbsp; &nbsp;\n" /* separator of menu items */
#define SMALLSPACE "<font size=\"-2\">&nbsp;</font>" /* good for \, */
//this is too long:
//#define SMALLSPACE "&thinsp;" /* good for \, */
#define MIDSPACE "<font size=\"-1\">&nbsp;</font>"   /* good for \; */
#define DOTS "&hellip;"
//#define DOTS "." SMALLSPACE "." SMALLSPACE "."
#define xDEBUG
#define INDENT 40 /* pixels the browser indents <blockquote> or <ul> etc. */
#define PPMSCALE "-g1" /* ppmscale option (-r -gGAMMA) 
                          NB: standard gamma=2.2 gives too light font */

#define eq !strcmp

FILE *q;

typedef unsigned char uchar;

typedef struct tok_s {
  struct tok_s *next,*prev;
  uchar *t; } tok_t;

char *strincl(char *dest,const char *incl)
/*
  includes string incl to dest before string dest
  example: dest="DEST", incl="INCL" ==> "INCLDEST"
  returns the pointer after incl (=the old contents of dest)
  does not check whethere there is memory enough in dest
  strings may not overlap
*/
{
  int l=strlen(incl);
  char *d,*i;

  for (d=dest+strlen(dest),i=d+l; d>=dest; d--,i--) *i=*d;
  memcpy(dest,incl,l);

  return dest+l;
}

int specchar(uchar *name)
{
  return eq(name,"\\{") || eq(name,"\\}") || eq(name,"\\$") || eq(name,"\\_") || eq(name,"\\^") || eq(name,"\\%") || eq(name,"\\#");
}

void debug(tok_t *t)
{
  fprintf(stderr,"\nDEBUG:%s:%s:%s:%s\n",
          t->t,t->next->t,t->next->next->t,t->next->next->next->t);
}

tok_t *head,*end;  

typedef struct macro_s {
  struct macro_s *next;
  int n;
  int LH; /* \LHnewcommand macros, not propagated to TeX */
  uchar *name;
  tok_t *body; } macro_t;

macro_t *macro0;

#define TOKENLEN 256 /* max token or identifier length */
uchar tok[TOKENLEN]=" ";

char *afterspace(tok_t *t)
{
  if (t->t[0]!='\\') return "";
  if (!t->t[1]) return "";
  if (isalpha(t->t[1]) && t->next && isalpha(t->next->t[0])) return " ";
  return "";
}

int warnonly=0; /* ugly patch: next call to tokerror is warning only */

void tokerror(tok_t *where,char *msg)
// error message
{
  tok_t *t,*w=where;
  int n;

  if (w) loop (n,0,10) if (w->next) w=w->next;

  fprintf(stderr,"ERROR %s:\n",msg);

  n=0;
  if (w) for (t=head; t && t!=w; t=t->next) {
    fprintf(stderr,"%s%s",t->t,afterspace(t));
    if (t==where) fprintf(stderr,"... ERROR %s\n",msg);
    if (n++>MAXITEMS) {
      tok_t *tt=t->next;
      n=1;
      while (t!=tt) {
        if (n++>MAXITEMS) {
          fprintf(stderr,"\nTOO MANY ITEMS\n");
          goto many; }
        tt=tt->next; }
      
      fprintf(stderr,"\nCYCLE IN DATA\n");
     many: break; } }

  fputc('\n',stderr);
  if (!warnonly) exit(2);
  warnonly=0;
}

void nexttok(void)
/*
  tok[] = { C | \C | \TOKEN | $ | $$ }
  %comments are removed
  returns tok[]="" on eof 
*/
{
  static int ch=' '; /* last char */
  uchar *t=tok;

 again:
  if (ch<0) *t=0;
  else if (ch=='$' || ch=='\'' || ch=='`') { /* $ $$ ' '' ` ``  */
    *t++=ch;
    ch=getc(q);
    if (ch==t[-1]) {
      *t++=ch;
      ch=getc(q); }
    goto end; }
  else if (ch=='-') { /* - -- --- */
    *t++=ch;
    ch=getc(q);
    if (ch=='-') {
      *t++=ch;
      ch=getc(q);
      if (ch=='-') {
        *t++=ch;
        ch=getc(q); } }
    goto end; }
  else if (ch=='\n') { /* \n\n\n.. -> \n\n */
    *t++=ch;
    ch=getc(q);
    if (ch=='\n') {
      *t++=ch;
      while (ch=='\n') ch=getc(q); }
    goto end; }
  else if (ch=='\t' || ch==' ') { /* make 1 white */
    *t++=' ';
    while (ch=='\t' || ch==' ') ch=getc(q);
    goto end; }
  else if (ch=='\\') {
    *t++=ch;
    ch=getc(q);
    if (isalpha(ch)) {
      while (ch>0 && isalpha(ch)) *t++=ch, ch=getc(q);
      /* remove all whites after \command */
      while (ch>=0 && (ch=='\t' || ch==' ')) ch=getc(q); }
    else {
      *t++=ch;
      ch=getc(q); }
    goto end; }
  else if (ch=='%') {
    do ch=getc(q); while (ch!='\n' && ch>=0);
    ch=getc(q);
    goto again; }
  else {
    *t++=ch;
    ch=getc(q); }

  end:
    *t=0;
/*.....    printf("token=\"%s\"\n",tok);*/
}

tok_t *findright(tok_t *left,uchar *l)
// find matching ] \] $ $$ } etc.
{
  int nest=0;
  tok_t *right;
  uchar *r=NULL; 

  if (eq(l,"{")) r="}";
  else if (eq(l,"[")) r="]";
  else if (eq(l,"\\[")) r="\\]";
  else if (eq(l,"\\(")) r="\\)";
  else if (eq(l,"$")) r="$";
  else if (eq(l,"$$")) r="$$";
  else tokerror(left,"internal: one of { [ \\[ \\( $ $$ expected");

/*.....    printf("---%s---%s---\n",left->t,l);*/

  if (strcmp(left->t,l)) tokerror(left,"left par expected");
  
  looplist (right,left->next)
    if (eq(right->t,r) && nest==0) return right;
    else if (eq(right->t,"}")) nest--;
    else if (eq(right->t,"{")) nest++;

  tokerror(left,"no closing parenthesis");
  return NULL;
}

tok_t *cut(tok_t *a,tok_t *b)
/* remove a->...->b incl., returns 1st item not removed */
{
  if (!a->prev) tokerror(a,"cut: no prev");
  if (!b->next) tokerror(b,"cut: no next");
  a->prev->next=b->next;
  b->next->prev=a->prev;

  return b->next;
}

tok_t *include(tok_t *a,tok_t *b)
/* include copy of b after a, returns copy of b */ 
{
  static int check;
  tok_t *n=a->next,*B;

  if (check++>MAXITEMS) tokerror(a,"too many includes (macro loop?)");

  allocone(B);
  B->t=mystrdup(b->t);

  a->next=B;
  B->next=n;
  B->prev=a;
  n->prev=B;

  return B;
}

static tok_t *from,*frombody,*after,*afterbody;
int ntok;

uchar *getgroup(tok_t *n)
{
#define GGLEN 12400
  static uchar name[GGLEN];
  tok_t *t;

/*.....  printf("getgroup: /%s/%s/\n",n->t,n->next->t);*/

  if (eq(n->t,"{")) {
    from=n; frombody=n->next;
    after=findright(n,"{");
    name[0]=0;
    for (t=frombody; t!=after; t=t->next)
      if (strlen(name)+strlen(t->t)<GGLEN) strcat(name,t->t+specchar(t->t));
    for (ntok=-1; n!=after; n=n->next) ntok++;
    afterbody=after;
    after=after->next; }
  else {
    from=frombody=n;
    strcpy(name,n->t);
    ntok=1;
    after=afterbody=n->next; }

  return name;
}

char *space(int h,tok_t *t)
/* h: 1=hspace, 0=vspace */
{
  char *n;
  static char line[LINELEN];
  double q=1;
  int i;

  if (eq(t->next->t,"*")) t=t->next;
  n=getgroup(t->next);
  if (strstr(n,"cm")) q=8;
  else if (strstr(n,"mm")) q=0.8;
  else if (strstr(n,"in")) q=20;
  else if (strstr(n,"pt")) q=20/72.;
  else if (strstr(n,"em")) q=2;
  else if (strstr(n,"ex")) q=1;
  else tokerror(t,"unsupported length unit");

  line[0]=0;
  if (q>200) q=200;
  q*=atof(n);

  if (h) {
    /* q is roughly in ex */
    loop (i,0,(int)q) strcat(line,"&nbsp;");
    if (q-(int)q>0.5) strcat(line,"&thinsp;"); }
  else {
    strcat(line,"<br>&nbsp;<br>");
    q=q/3-1;
    loop (i,0,(int)q) strcat(line,"&nbsp;<br>"); }
    
  return line;
}

#include "section.c"

section_t lastsection;

typedef struct xref_s {
  struct xref_s *next;
  uchar *name;
  uchar *latexlabel;
  section_t section; } xref_t;

int ntoc,isshortentop=1;

xref_t *xref0;

xref_t *findxref(uchar *s)
{
  xref_t *x;

  looplist (x,xref0) if (eq(x->name,s)) return x;

  fprintf(stderr,"WARNING %s undefined label\n",s);
  return NULL;
}

int beginend(tok_t *t,char *env,char *be)
{
  if (eq(t->t,be)) {
    t=t->next;
    if (strcmp(t->t,"{")) tokerror(t,"missing { after \\begin or \\end");
    t=t->next;
    return eq(t->t,env); }
  return 0;
}

int isbegin(tok_t *t,char *env)
/* returns true if \begin{env} found, merged tokens assumed */
{
  return beginend(t,env,"\\begin");
}

int isend(tok_t *t,char *env)
/* returns true if \end{env} found, merged tokens assumed */
{
  return beginend(t,env,"\\end");
}


tok_t *findend(tok_t *t)
/* find } of matching \end{}; t -> \begin */
{
  uchar *env=t->next->next->t;

  if (strcmp(t->t,"\\begin") || strcmp(t->next->t,"{")) tokerror(t,"\\begin{ expected");

#ifdef DEBUG
fprintf(stderr,"//////%s//////\n",env);
#endif

  for (t=t->next->next; t; t=t->next)
    if (eq(t->t,"\\end") && eq(t->next->next->t,env)) return t->next->next->next->next;

  tokerror(t,"\\end not found");
  return NULL;
}

FILE *out,*outhtml,*outaux;

char *sectionlevel(char *key,section_t section)
/*
  generates data for l2h2 (2nd pass of LaTeX->HTML translation)
  key in l2h.aux:
    ITEM : section name
  key in html file generated by l2h:
    LONGITEM : marks place where the (unabridged) name of section will appear
    SHORTITEM : marks place where the shortened name of section will appear
    MENU : marks place where the menu will appear
*/
{
  static char line[128];

  sprintf(line,"<!--%s %d %d %d %d %d\n",
          key,section.level,
          section.s[0],section.s[1],section.s[2],section.s[3]);

  return line;
}

char *topright(void)
{
  static char *tr;

  if (!tr) {
    FILE *f=fopen("topright.html","rb");
    char line[LINELEN];
    int n=0;

    if (!f) Error("topright.html");
    while (fgets(line,LINELEN,f)) n+=strlen(line);
    alloc(tr,n+1);
    tr[0]=0;
    rewind(f);
    while (fgets(line,LINELEN,f)) strcat(tr,line); }

  return tr;
}

void openhtml(char *fn,section_t section)
{
  char *hn;

  fn=mystrdup(fn);

  ntoc++;
  lastsection=section;

  out=outhtml=fopen(fn,"wt");
  fprintf(stderr,"writing %s (ntoc=%d, level=%d)\n",fn,ntoc,section.level);
  fprintf(out,"\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n\
<html>\n\
<head>\n\
<META HTTP-EQUIV=\"Content-Type\" content=\"text/html; charset=iso-8859-2\">\n\
<link rel=\"stylesheet\" type=\"text/css\" href=\"prikladnik.css\">\n\
<script src=\"script/appear.js\" type=\"text/javascript\"></script>\n");
  if (soundconfig) fprintf(out,"\
<script src=\"script/soundmanager.js\" type=\"text/javascript\"></script>\n");
  fprintf(out,"\
<TITLE>%s%s</TITLE>\n\
</head>\n\
<body onload=\"displayinit();\"%s\n\
<DIV id=\"main\"><div><div><div>\n\
<Div id=\"top\">\n\
<div id=\"topright\"><ul>\n%s\
</ul></div><!--id=\"topright\"-->\n",
	  sectionlevel("TITLE",section),"-->\n",
	  soundconfig?" onkeypress=\"showKeyPress(event)\"><div></div>":">",
	  topright());

  /* top navigation panel: section navigation */

  checkhtmlname=1;

  fprintf(out,"\n<div id=\"topleft\">\n");

  /* UP */
  if (section.level == SECTION) {
    section.level--;
    if ( (hn=htmlname(section)) ) fprintf(out,"\
<a href=\"%s\" class=\"up\" title=\"nahoru\"> %s-->\n</a>"MENUSEP,hn,sectionlevel(isshortentop?"SHORTITEM":"LONGITEM",section));
    section.level++; }

  /* UPUP */
  else if (section.level == SUBSECTION) {
    section.level-=2;
    if ( (hn=htmlname(section)) ) fprintf(out,"\
<a href=\"%s\" class=\"upup\" title=\"o 2 rovn nahoru\"> %s-->\n</a>"MENUSEP,hn,sectionlevel(isshortentop?"SHORTITEM":"LONGITEM",section));
    section.level++; 

    if ( (hn=htmlname(section)) ) fprintf(out,"\
<a href=\"%s\" class=\"up\" title=\"nahoru\"> %s-->\n</a>"MENUSEP,hn,sectionlevel(isshortentop?"SHORTITEM":"LONGITEM",section));
    section.level++; }
  else
    if (section.level!=BOOK) 
      fprintf(out,"<a href=\"%shtml\" class=\"longskripta\" title=\"vod, obsah\">Obsah</a>",fileprefix);

  fprintf(out,"</div><!--topleft-->\n<div id=\"topbottom\">\n");

  if (section.level!=BOOK) {

    /* PREV */
    section.s[section.level]--;
    if ( (hn=htmlname(section)) ) fprintf(out,"\
<a href=\"%s\" class=\"prev\" title=\"pedchoz\">%s-->\n</a>"MENUSEP,hn,sectionlevel("SHORTITEM",section));
    section.s[section.level]++;
    
    /* NEXT */
    section.s[section.level]++;
    if ( (hn=htmlname(section)) ) fprintf(out,"\
<a href=\"%s\" class=\"next\" title=\"nsledujc\">%s-->\n\
</a>\n",hn,sectionlevel("SHORTITEM",section));
    section.s[section.level]--; }

  fprintf(out,"</div><!--topbottom-->\n\
</Div><!--top-->\n\
<Div id=\"middle\">");

  free(fn);
  checkhtmlname=0;
}

void makemenu(section_t section)
{

  /* menu header text */
  switch (section.level) {
    case BOOK: fputs("<h3>Obsah:</h3>\n",out); break;
    case CHAPTER: fputs("<h3>Tmata:</h3>\n",out); break;
    case SECTION: fputs("<h3>Pklady:</h3>\n",out); break;
    case SUBSECTION: fputs("<a name=\"avpr\"></a> <ul id=\"vpr\">\n",out); break;
    case SUBSUBSECTION: ; }

  /* mark menu for second pass */
  fprintf(out,sectionlevel("MENU",section));
  fprintf(out,"-->\n");
}

void closehtml(void)
{
  fprintf(out,"\n</DIV>\n");
      
  if (lastsection.level == SUBSECTION)
    fprintf(out,"<div id=\"downjs\"><p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;<p>&darr;</div>\n");

  fprintf(out,"</Div  >\n\
<Div id=\"bottom\">");

  makemenu(lastsection);

  if (out!=outhtml) Error("bad output file (section structure corrupted)");

  if (lastsection.level == SUBSECTION) fprintf(out,"</ul>\n");

  fprintf(out,"\
</Div><!--id=\"bottom\"-->\n\
</div></div></div></DIV><!--id=\"main\"-->\n");
  if (soundconfig)
    fprintf(out,"<script type=\"text/javascript\">soundManagerInit();</script>\n");
  fprintf(out,"</body></html>\n");
  fclose(out);
  out=outhtml=NULL;
}

void reopenhtml(char *fn,section_t section)
{
  if (out) closehtml();
  openhtml(fn,section);
}

/*
  zarizuje, aby se pro dve ruzna pouziti tehoz $..$ pouzil jeden obrazek 
  neni 100%: dva vyrazy jsou stejne, maji-li stejne CRC a delku
  dale zarizuje znovupouziti jiz jednou udelanych obrazku v opravenem behu
*/
struct mathtex_s {
  struct mathtex_s *next;
  unsigned4 crc;
  int nchar,size;
  int count; /* kolikrat pouzito */
  int lchap,leqno; /* for equation, eqnarray: numbering; lchap=0: simple math */
  char *html; /* html tag */
} *mathtex0,*matx0;

typedef struct stat_s {
  enum font_e { RM,IT,TT,SF } font;
  int underline;
  int bold;
  int size;
  int nie;
  int ie[8]; /* 0=end, +1=superscript, -1=subscript */
} stat_t;

#define SUP  1
#define SUB -1

stat_t stat;

char *fontcolor=""; /* for chapter... only */

char *fontchange(stat_t stat,int theend)
{
  static char line[256];
  int nie;
  line[0]=0;

  if (theend) {
    if (stat.font==SF) strcat(line,"</font>");
    if (stat.font==TT) strcat(line,"</tt>");
    if (stat.font==IT) strcat(line,"</i>");
    if (stat.bold) strcat(line,"</b>");
    if (stat.underline) strcat(line,"</u>");
    if (stat.size) strcat(line,"</font>");
    for (nie=stat.nie; nie>=0; nie--)
      if (stat.ie[nie]) strcat(line,stat.ie[nie]>0?"</sup>":"</sub>");
    }
  else {
    loop (nie,0,stat.nie) strcat(line,stat.ie[nie]>0?"<sup>":"<sub>");
    if (stat.size) strcat(line,string("<font %ssize=\"%+d\">",fontcolor,stat.size));
    if (stat.underline) strcat(line,"<u>");
    if (stat.bold) strcat(line,"<b>");
    if (stat.font==IT) strcat(line,"<i>");
    if (stat.font==TT) strcat(line,"<tt>");
    if (stat.font==SF) strcat(line,"<font face=\"helvetica,arial\">");
    }
  fontcolor="";
  return line;
}

#define BLOCKQUOTE 1 /* must be 1 */
/* ... To fix a bug in MSIE 5.0, BLOCKQUOTE was changed to TABLE */
#define LEFT 2
#define MIDDLE 4
#define RIGHT 8
#define BOTTOM 16
#define POS 32

FILE *openedimages,*cropinfo;
char ctline[LINELEN],testr[256];
int indent=0;
struct list_s {
  int n;
  char type; 
} list;

char *imginfo(char *imgname,int place,int pos,char *alt)
/* GIF: reads and analyzes file, JPEG: calls jpegtopnm */
{
  char *ext=imgname+strlen(imgname)-4;
  static char line[256];
  int width=0,height=0;

  fprintf(stderr,"reading %s\n",imgname);
  if (openedimages) fprintf(openedimages,"%s",imgname);
  if (cropinfo) fprintf(cropinfo,"%s:%s%s\n",imgname,testr,ctline);

  if (!strcmp(ext,".jpg")) {
    FILE *ppm;

    if (system(string("rm /tmp/jpeg.ppm; jpegtopnm %s > /tmp/jpeg.ppm",imgname)))
      Error("cannot convert JPEG - jpegtopnm missing?");

    ppm=fopen("/tmp/jpeg.ppm","r");
    fgets(line,256,ppm);
    if (line[0]!='P') Error("ppm expected");
    do fgets(line,256,ppm); while (line[0]=='#');
    sscanf(line,"%d%d",&width,&height); 
    fclose(ppm); }

  else if (!strcmp(ext,".gif")) {
    FILE *gif=fopen(imgname,"rb");
    short int hdr[5];
    
    if (!gif) Error("cannot open GIF");

    if (5!=fread(hdr,2,5,gif)) Error("read GIF");
    fclose(gif);
    if (memcmp(hdr,"GIF8",4)) Error("GIF8.. expected");
    width=hdr[3]; height=hdr[4]; }
  else
    Error("unsupported image file");

  /* used also in math.lst => no \n inside line */
  sprintf(line,"%s<img src=\"%s\"%s border=0 alt=\"%s\"",
	  place&BLOCKQUOTE?"<p>":"",
          imgname,
          place&LEFT?" align=left":
            place&BOTTOM?" align=bottom":
            place&MIDDLE?" align=middle":
            place&RIGHT?" align=right":
            place&POS?string(" class=\"pos%d\"",pos):
	    place&BLOCKQUOTE?" class=\"dsp\"":"",
	  alt);
  
  sprintf(strend(line)," width=%d height=%d",width,height);

/*.....  if (openedimages) fprintf(openedimages," %d %d",width,height);*/
  if (openedimages) fprintf(openedimages,"\n");

  strcat(line,">");
  if (place&BLOCKQUOTE) strcat(line,"<p>\n");

  return line;
}

struct ignored_s {
  struct ignored_s *next;
  char *name; } *ignored0;

int eqtag(uchar *a,uchar *b)
/* a and b start at tag name, not at < */
{
  for ( ; *a && *b; a++,b++) {
    if (*a=='>' && *b=='>') return 1;
    if (*a!=*b) return 0; }
  Error("eqtag");
}

void removetag(uchar *t)
{
  uchar *e=strchr(t,'>');
  
  if (!e) Error("removetag");
  e++;
  if (*--t=='/') t--;
  memmove(t,e,strlen(e)+1);
}

int htmltable=0; /* nesting of htmltable+longtable */
struct {
  int n,i; /* columns, counter & */
  int border; /* at least 1 | causes border + bg colors */
  int bgminus1; /* = bg[-1] (dirty!!!) */
  int bg[32];
  char *align[32]; /* max 32 columns */
} tab[4]; /* max 4 nested (WARNING: htmltable/ongtable/tabular?) */
int emptycol; /* BUG: this should be in tab[] because of nesting!!!*/
int titlerow=0; /* BUG: this should be in tab[] because of nesting!!!*/

void htmltabtd(void)
{
  int itab=tab[htmltable-1].i++;

  fprintf(out,"<td align=%s bgcolor=%s>",
	  tab[htmltable-1].align[itab],
	  titlerow ? 
	    tab[htmltable-1].bg[itab]<2?"\"#cfffff\"":"\"#cfffcf\""
	           :
	    tab[htmltable-1].bg[itab]<2?"white":"\"#ffffcf\""
	  );
}


char greeksymbol[256];

static struct gr_s { char *latex,*html; } gr[]={
  { "\\AA","&Aring;" },
  { "\\alpha","&alpha;" },
  { "\\beta","&beta;" },
  { "\\gamma","&gamma;" },
  { "\\delta","&delta;" },
  { "\\epsilon","&#1013;" },
  { "\\varepsilon","&epsilon;" },
  { "\\zeta","&zeta;" },
  { "\\eta","&eta;" },
  { "\\theta","&theta;" },
  { "\\vartheta","&thetasym;" },
  { "\\iota","&iota;" },
  { "\\kappa","&kappa;" },
  { "\\lambda","&lambda;" },
  { "\\mu","&mu;" },
  { "\\nu","&nu;" },  // LOOKS BAD
  { "\\xi","&xi;" },
  { "\\pi","&pi;" },
  { "\\rho","&rho;" },
  { "\\varrho","&#1009;" },
  { "\\sigma","&sigma;" },
  { "\\tau","&tau;" },
  { "\\upsilon","&upsilon;" },
  { "\\phi","&#981;" },
  { "\\varphi","&phi;" },
  { "\\chi","&chi;" },  // LOOKS BAD
  { "\\psi","&psi;" },
  { "\\omega","&omega;" },
  { "\\vartheta","&vartheta;" },
  { "\\varpi","&varpi;" },
  { "\\varphi","&varphi;" },
  { "\\Gamma","&Gamma;" },
  { "\\Delta","</i>&Delta;<i>" }, // too inclined
  { "\\Theta","&Theta;" },
  { "\\Lambda","&Lambda;" },
  { "\\Xi","&Xi;" },
  { "\\Pi","&Pi;" },
  { "\\Sigma","&Sigma;" },
  { "\\Upsilon","&Upsilon;" },
  { "\\Phi","&Phi;" },
  { "\\Psi","&Psi;" },
  { "\\Omega","&Omega;" },
  { "\\partial","&part;" },
  { "\\sum","&sum;" },
  { "\\int","&int;" },
  { "\\infty","&infin;" },
  { "\\ne",EQSEP"&ne;"EQSEP },
  { "\\le",EQSEP"&le;"EQSEP },
  { "\\ge",EQSEP"&ge;"EQSEP },
  { "\\approx",EQSEP"&asymp;"EQSEP },
  { "\\sim",EQSEP"&sim;"EQSEP },
  { "\\langle","&lang;" },
  { "\\rangle","&rang;" },
  { "\\equiv","&equiv;" },
  { "\\bullet","&bull;" },
  { "\\to",EQSEP"&rarr;"EQSEP },
  { "\\rightarrow",EQSEP"&rarr;"EQSEP },
  { "\\longrightarrow",EQSEP"&#10230;"EQSEP },
  { "\\Rightarrow",EQSEP"&rArr;"EQSEP },
  { "\\leftarrow",EQSEP"&larr;"EQSEP },
  { NULL,NULL} };

int isgreek(char *t)
     /* greek and other symbols */
{
  int i;

  for (i=0; gr[i].latex; i++) if (eq(t,gr[i].latex)) {
    strcpy(greeksymbol,gr[i].html);
    return 1; }

  return 0;
}

int fileexists(char *fn)
{
  FILE *f=fopen(fn,"r");
  if (f) {
    fclose(f);
    return 1; }
  else
    return 0;
}

int main(int narg,char **arg)
{
  tok_t *t;
  static int isdebug,isgraphmath,iszip,isrerun,ignoremathlst;
  static double belowex=-3,aboveex=3;
  static int shorts=44,ppmcolor=0;
  int flag,nimg=0;
  static char *fn;
  int nest=0,iarg;
  int size=SIZE;
  int bottom=0; /* math img placing */
  double fshift=3.5;
  double scale=1; /* additional scaling of generated GIFs */
  int finalres;
  int yshift;
  static stat_t stat[MAXNEST];
  section_t section={NULL,BOOK,{0,0,0,0}};

  if (narg<2) {
    fprintf(stderr,"Call by:\n\
  l2h [OPTIONS] NAME [OPTIONS]\n\
Processes files NAME.tex and NAME.aux, generates various html and gif files\n\
OPTIONS:\n\
  -a#\tspace above line in $..$->GIF conversion, in ex [default=%.2f]\n\
  -b#\tspace below line, in ex [default=%.2f], normally minus the value of -a\n\
  -c\tdo not shorten chapter and subsection names for top bar [default=all]\n\
  -C\trender all math + EPS in color (gs DEVICE=ppmraw) [default=pgmraw=gray]\n\
  -D\tdebug - don't generate GIFs, just write img name instead\n\
  -d#\tDPI for LaTeX [12pt] to GIF conversion [default=%d]\n\
  -f\taudio support via Adobe Flash (formerly ShockWave or Macromedia Flash)\n\
  -g\tdoes not assume that the browser knows Greek alphabet; ignored if -m\n\
  -i\tignore math.lst and generate all GIFs again (USE IF -d,-x,-C CHANGED!)\n\
  -k#\tshift of baseline w. ind/exp for the MSIE 5.5 version [default=%.2f]\n\
  -m\tdo all math graphically [default=translate $simple math$ into HTML]\n\
  -x#\tantialiasing in square #*# [default=%d]\n\
  -s#\tshorten subsubsection names to # chars (exported to l2h2) [default=%d]\n\
  -t@\tPGM to GIF filter [default=%s]\n\
  -z\tzip all index*html (before running l2h2 on them)\n\
",aboveex,belowex,DPI,fshift,SCALE,shorts,ppmtogif);
    exit(0); }

  loop (iarg,1,narg) if (arg[iarg][0])
    fprintf(stderr,"arg[%d]=\"%s\"\n",iarg,arg[iarg]);

  loop (iarg,1,narg) if (arg[iarg][0]) {
    if (arg[iarg][0]=='-') switch (arg[iarg][1]) {
      case 'D': isdebug++; break;
      case 'm': isgraphmath++; break;
      case 'i': ignoremathlst++; break;
      case 'r': Error("RERUN no longer supported");
      case 'a': aboveex=atof(arg[iarg]+2); break;
      case 'd': DPI=atoi(arg[iarg]+2); break;
      case 'g': gr[0].latex=gr[0].html=NULL; break;
      case 'f': soundconfig=fopen("sound-config.xml","wt"); break;
      case 'x': SCALE=atoi(arg[iarg]+2); break;
      case 's': shorts=atoi(arg[iarg]+2); break;
      case 'b': belowex=atof(arg[iarg]+2); break;
      case 't': ppmtogif=arg[iarg]+2; break;
      case 'k': fshift=atof(arg[iarg]+2); break;
      case 'c': isshortentop=0; break;
      case 'C': ppmcolor++; ppmtogif=colorppmtogif; break;
      case 'z': iszip++; break;
      default: Error("l2h: bad option"); }
    else fn=arg[iarg]; }

  SIZE=DPI*SCALE;
  yshift=(int)(fshift*SCALE+.5);

  openedimages=fopen("images.lst","wt");
  
  if (soundconfig)
    fprintf(soundconfig,"\
<!-- SoundManager Flash configuration -->\n\
<!-- Sound IDs are referenced from JS -->\n\
\n\
<items baseHref=\"audio/\" defaultVol=\"100\">\n");

  if (isdebug && isrerun) Error("cannot combine -d and -r");
  if (aboveex<=0) Error("negative -a");
  if (belowex>=0) Error("positive -b");
  stringinit(LINELEN,16);

  fileprefix[0]='p';
  strcat(fileprefix,".");

  if (!fn) Error("no NAME");
  q=fopen(string("%s.tex",fn),"rt");
  if (!q) Error(fn);

  /* cte informace o obrazcich */
  if (!isdebug && !isrerun) {
    FILE *mtf;
    struct mathtex_s *mt;
    char line[LINELEN],html[LINELEN*10];

    cropinfo=fopen("cropinfo.aux","wt");

    if (ignoremathlst) mtf=NULL;
    else mtf=fopen("math.lst","rt");

    if (!mtf)
      fprintf(stderr,">>>>> no or ignored math.lst <<<<<\n");
    else {
      fprintf(stderr,">>>>> reading math.lst <<<<<\n");
      fgets(line,LINELEN,mtf);
      
      while (fgets(line,LINELEN,mtf)) if (strlen(line)>10) {
        alloconezero(mt);
        mt->next=matx0; matx0=mt;
        sscanf(line,"%x %d %d %d %d", &mt->crc,&mt->nchar,&mt->size,&mt->lchap,&mt->leqno);
        html[0]=0;
        do {
          if (!fgets(line,LINELEN,mtf)) Error("math.lst format");
          if (strlen(html)>LINELEN*10-LINELEN) Error("math.lst long line");
          strcat(html,line);
          } while (line[0]!='\n');
        mt->html=mystrdup(html); }

      fclose(mtf); }
    }

  /* allocating the list of tokens */
  for (; tok[0]; nexttok()) {
    alloconezero(t);
    t->t=mystrdup(tok);
    t->prev=end; end=t; }

  /* making ->next list */
  for (t=end; t; t=t->prev) {
    head=t;
    if (t->prev) t->prev->next=t; }

  /* test print */
/*.....  looplist (t,head) printf("\"%s\"\n",t->t);*/

  /* parsing macros - only \[LH][re]newcommand{\name}{body} allowed */
  looplist (t,head) if (eq(t->t,"\\newcommand") || eq(t->t,"\\renewcommand") || eq(t->t,"\\LHnewcommand")) {
    uchar *name;
    tok_t *n=t->next;
    int N=0;
    int re=eq(t->t,"\\renewcommand");
    int LH=eq(t->t,"\\LHnewcommand");
    macro_t *m;

    name=getgroup(n);
    if (ntok!=1) tokerror(n,"simple name in \\newcommand expected");
    n=after;

    looplist (m,macro0)
      if (eq(m->name,name)) {
        if (!re) {
          warnonly=1;
          tokerror(t,"macro redefined (will cause LaTeX warnings later)"); 
          re=1; }
        goto toredefine; }
    if (re) {
      warnonly=1;                                                           
      tokerror(t,"no such macro to redefine (will cause LaTeX warnings later)");
      re=0; }

   toredefine:
 
    if (eq(n->t,"[")) {
      n=n->next;
      N=atoi(n->t);
      if (N>9) tokerror(n,"more than 9 macro args not supported");
      n=n->next;
      if (strcmp(n->t,"]")) tokerror(n,"] missing in \newcommand{\\NAME}[N]"); 
      n=n->next; 
      if (n->t[0]<=' ') n=n->next; }

    if (!m) {
      alloconezero(m);
      m->name=mystrdup(name);
      m->LH=LH;
      m->next=macro0;
      macro0=m; }
    m->n=N;
    m->body=n->next;
    if (eq(m->body->t,"}")) {
      /* hook for \newcommand\MACRO{} */
      m->body=NULL;
      n=n->next; }
    else
      n=findright(n,"{");
    n->prev->next=NULL;
    t=cut(t,n); }

  /* fake defining LHunitchem */
  {
    macro_t *m;
    alloconezero(m);
    m->name="\\LHunitchem";
    m->LH=2;
    m->next=macro0;
    macro0=m;
    m->n=2;
    m->body=NULL; }

/*.....  { macro_t *m; looplist (m,macro0) fprintf(stderr,"MACRO  %s[%d] %d\n",m->name,m->n,m->LH); }*/

  /* expanding macros */
  flag=1;
  while (flag) {
    flag=0;
    looplist (t,head) {
      macro_t *m;

      looplist (m,macro0) if (eq(m->name,t->t)) {
        int i;
        tok_t **arg=NULL,*n=t->next;
        
        flag++;
        
        if (m->n) {
          allocarray(arg,m->n);
          loop (i,0,m->n) {
            getgroup(n);
            arg[i]=frombody;
            n=after;
            /* ignore whitespaces between macro arguments (not after the last) */
            if (i!=m->n-1)
              while (strlen(n->t)==1 && strchr(" \n\t\r",n->t[0])) n=n->next;
            if (frombody==afterbody) 
              arg[i]=NULL; /* empty arg */
            else {
              cut(from,after->prev);
              if (afterbody==after) frombody->next=NULL; /* \macro X */
              else afterbody->prev->next=NULL; } } }     /* \macro{X} */
        
        /* getting rid of the macro name */
        t=cut(t,t)->prev;

        /* replacing args */
        if (m->LH==2) { /* LHunitchem */
          tok_t *x;
          static tok_t dol={NULL,NULL,"$"};
          static tok_t up={NULL,NULL,"^"};
          static tok_t dn={NULL,NULL,"_"};
          static tok_t beg={NULL,NULL,"{"};
          static tok_t end={NULL,NULL,"}"};
          static tok_t spc={NULL,NULL,"\\,"};
          static tok_t cdot={NULL,NULL,"\\cdot"};
          static tok_t hyph={NULL,NULL,"-"};
          int nummode=0;
          uchar key=arg[0]->t[0];
          tok_t *updn=isupper(key)?&up:&dn;

          looplist (x,arg[1]) {
            if (strchr("0123456789+-",x->t[0])) {
              if (!nummode) {
                t=include(t,&dol);
                t=include(t,updn);
                t=include(t,&beg); }
              nummode=1;
              t=include(t,x); }
            else {
              if (nummode) {
                t=include(t,&end);
                t=include(t,&dol); }
              nummode=0;
              if (eq(".",x->t)) t=include(t,&spc);
              else if (eq("\'",x->t)) {
                t=include(t,&dol);
                t=include(t,&cdot);
                t=include(t,&dol); }
              else if (eq("\\-",x->t)) t=include(t,&hyph);
              else t=include(t,x); } }
          if (nummode) {
            t=include(t,&end);
            t=include(t,&dol); }
          }

        else looplist (n,m->body) /* general macro */
          if (eq(n->t,"#")) {
            int i;
            tok_t *x;
          
            n=n->next;
            i=atoi(n->t)-1;
            if (i>=m->n) tokerror(n,"# with invalid arg");
          
            looplist (x,arg[i]) t=include(t,x); }
          else
            t=include(t,n);
        
        if (arg) free(arg); }
      } }


  /* merge tokens in cases like \begin{xxx}: xxx will be one token  */
  looplist (t,head)
    if (eq(t->t,"\\begin") || eq(t->t,"\\end")
     || eq(t->t,"\\label") || eq(t->t,"\\ref")) {
      char *name=getgroup(t->next);

      if (strcmp(t->next->t,"{")) tokerror(t,"{ expected");

      frombody->t=mystrdup(name);
      frombody->next=afterbody;
      afterbody->prev=frombody; }

  /* expanding _x into _{x} and ^x into ^{x} (for easier further parsing) */
  looplist (t,head)
    if (eq(t->t,"_") || eq(t->t,"^")) {
      if (t->next->t[0]<=' ') tokerror(t,"implementation limitation: space after _ or ^ not allowed");
      if (strcmp(t->next->t,"{")) {
        tok_t *n;

        allocone(n);
        n->t="{";
        n->prev=t;
        n->next=t->next;
        t->next->prev=n;
        t->next=n;
        t=n->next;

        allocone(n);
        n->t="}";
        n->prev=t;
        n->next=t->next;
        t->next->prev=n;
        t->next=n; } }


/*.....  looplist (t,head) printf("\"%s\"\n",t->t);*/
#if 0
  printf("=======================\n");
  looplist (t,head) printf("_%s_",t->t);  
  printf("=======================\n");
#endif

  /* parsing \label{} statements */
  section.s[0]=0;
  looplist (t,head) {
    if (eq(t->t,"\\chapter") || eq(t->t,"\\section") || eq(t->t,"\\subsection") || eq(t->t,"\\subsubsection"))
      tokerror(t,"\\chapter{} etc. not allowed, use macros in l2h.tex to replace");

    if (eq(t->t,"\\LHchapter")) {
      section.s[0]++,section.s[1]=0,section.level=CHAPTER;
      if (indent) tokerror(t,"chapter inside quote or list"); }
    if (eq(t->t,"\\LHsection")) {
      section.s[1]++,section.s[2]=0,section.level=SECTION;
      if (indent) tokerror(t,"section inside quote or list"); }
    if (eq(t->t,"\\LHsubsection")) {
      section.s[2]++,section.s[3]=0,section.level=SUBSECTION;
      if (indent) tokerror(t,"subsection inside quote or list"); }
    if (eq(t->t,"\\LHsubsubsection")) {
      section.s[3]++,section.level=SUBSUBSECTION;
      if (indent) tokerror(t,"subsubsection inside quote or list"); }

    if (eq(t->t,"\\label")) {
      xref_t *x,*y;

      alloconezero(x);
      x->next=xref0; xref0=x;
      x->section=section;
      x->name=t->next->next->t;
      looplist (y,x->next) if (!strcmp(x->name,y->name))
        tokerror(t,"label not unique"); } }

  /* read .aux to get LaTeX xref */
  section.s[0]=0;
  section.level=BOOK;
  ntoc=0;
  {
    FILE *f=fopen(string("%s.aux",fn),"rt");
    uchar line[4096],*latexname,*latexlabel,*c;
/*.....    toc_t *end=NULL,*toc;*/

    if (!f) Error("no LaTeX aux file - run latex first!");
    while (fgets(line,4096,f)) {

      if (!memcmp(line,"\\newlabel{",10)) {
        xref_t *x;

        c=strchr(line+10,'}');
        if (!c) Error("LaTeX aux file: missing }");
        *c=0;
        latexname=line+10;
        latexlabel=c+3;
        c=strchr(latexlabel,'}');
        if (!c) Error("LaTeX aux file: missing }");
        *c=0;
        
        looplist(x,xref0)
          if (eq(x->name,latexname)) {
            x->latexlabel=mystrdup(latexlabel);
            x->section=section; /*?*/
            goto latexlabelfound; }

        fprintf(stderr,"latexname=%s\n",latexname);
        Error("latexlabel not found - have you re-run LaTeX?");
      latexlabelfound:; }

      if (!memcmp(line,"\\@writefile{toc}{\\contentsline {",32)) {
        uchar *c;
        section_t *s;

        ntoc++;

        c=line+32;
        if (!memcmp("chapter",c,7)) section.level=CHAPTER;
        else if (!memcmp("section",c,7)) section.level=SECTION;
        else if (!memcmp("subsection",c,10)) section.level=SUBSECTION;
        else if (!memcmp("subsubsection",c,13)) section.level=SUBSUBSECTION;
        else Error(c);

        section.s[section.level]++;
        if (section.level<SUBSUBSECTION) section.s[section.level+1]=0;

        allocone(s);
        *s=section;
        s->next=sec0; sec0=s; }

      }
    fclose(f);
    } /* read aux */


  /* generating output */
  section.s[0]=0;
  section.level=BOOK;
  ntoc=-1;
  outaux=fopen("l2h.aux","wt");
  openhtml(htmlname(section),section);


  looplist (t,head) if (isbegin(t,"document")) break;

  if (!t) Error("no \\begin{document}");

  /* <<<<< */
  for (t=t->next->next->next->next; t->next->next; t=t->next) {
    int simplemath=eq(t->t,"\\(") || eq(t->t,"$");
    int noenvmath=eq(t->t,"$$") || eq(t->t,"\\[");
    int epsfig=eq(t->t,"\\epsffile"); 
    int eqnarray=isbegin(t,"eqnarray");
    int numbermath=isbegin(t,"equation") || eqnarray;
    int envmath=numbermath || isbegin(t,"displaymath") || isbegin(t,"eqnarray*") || isbegin(t,"tabular");
    int display=noenvmath || envmath || epsfig;
    /* tabular --> GIF
       htmltable --> HTML table
       longtable --> HTML table */
/*.....    fprintf(stderr,"==%s==nest=%d\n",t->t,nest);*/

    if (t->t[0]>' ' && strcmp(t->t,"&")  && strcmp(t->t,"\\\\")) emptycol=0;

    if (isend(t,"document")) break;

    /* nesting of commands \bf etc. */
    else if (eq(t->t,"{") || eq(t->t,"\\begin")) {
      if (!envmath) {
        if (++nest>=MAXNEST) tokerror(t,"nesting exceeds MAXNEST");
        stat[nest]=stat[nest-1]; } }
    else if (eq(t->t,"}") || eq(t->t,"\\end")) {
      fprintf(out,fontchange(stat[nest],1));
      if (--nest<0) tokerror(t,"math nesting");
      fprintf(out,fontchange(stat[nest],0)); }

    if (eq(t->t,"\\rm")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].font=RM; }

    else if (eq(t->t,"\\sf")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].font=SF;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\tt")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].font=TT;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\it") || eq(t->t,"\\em")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].font=IT;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\bf")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].bold=1;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\LHul")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].underline=1;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\large")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].size=1;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\Large") || eq(t->t,"\\LARGE") || eq(t->t,"\\huge")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].size=3;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\small")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].size=-1;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\scriptsize") || eq(t->t,"\\tiny")) {
      fprintf(out,fontchange(stat[nest],1));
      stat[nest].size=-3;
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\label")) {
      t=t->next->next;
      fprintf(out,"<a name=\"%s\"></a>\n",t->t);
      t=t->next; }

    else if (eq(t->t,"\\ref")) {
      xref_t *x;
      char *hn;

      t=t->next->next;
      x=findxref(t->t);
      if (x) {
        hn=shorthtmlname(x->section);
        fprintf(out,"<a href=\"%s#%s\">%s</a>", hn,x->name,x->latexlabel); }
      else
        fprintf(out,"<a href=\"undefined.html\">UNDEFINED</a>");
      t=t->next; }

    /* ntoc is advanced by openhtml */
    else if (eq(t->t,"\\LHchapter")) {
      section.s[0]++; section.s[1]=0; section.level=CHAPTER;
      //      size=SIZE*5/3;
      size=SIZE;
      reopenhtml(htmlname(section),section);
      fprintf(out,fontchange(stat[nest],1));
      //      stat[nest].size=3; stat[nest].bold=1;
      //      fontcolor="color=\"#ff0000\" ";
      stat[nest].size=0; stat[nest].bold=0; fontcolor="";
      fprintf(out,fontchange(stat[nest],0)); 
      fprintf(out,"\n<H2>");
      fprintf(out,sectionnumbers(section)); }

    else if (eq(t->t,"\\LHendchapter")) {
      fprintf(out,"</H2>\n<DIV>\n"); }

    else if (eq(t->t,"\\LHsection")) {
      section.s[1]++; section.s[2]=0; section.level=SECTION;
      //      size=SIZE*4/3;
      size=SIZE;
      reopenhtml(htmlname(section),section);
      fprintf(out,fontchange(stat[nest],1));
      //      stat[nest].size=3; stat[nest].bold=0;
      //      fontcolor="color=\"#7f5500\" ";
      stat[nest].size=0; stat[nest].bold=0; fontcolor=""; // void->css
      fprintf(out,fontchange(stat[nest],0));
      fprintf(out,"\n<H3>");
      fprintf(out,sectionnumbers(section)); }

    else if (eq(t->t,"\\LHendsection")) {
      fprintf(out,"</H3>\n<DIV>\n"); }

    else if (eq(t->t,"\\LHsubsection")) {
      section.s[2]++; section.s[3]=0; section.level=SUBSECTION;
      //      size=SIZE*4/3;
      size=SIZE;
      reopenhtml(htmlname(section),section);
      fprintf(out,fontchange(stat[nest],1));
      //      stat[nest].size=1; stat[nest].bold=1;
      //      fontcolor="color=\"#00aa00\" ";
      stat[nest].size=0; stat[nest].bold=0; fontcolor=""; // void->css
      fprintf(out,fontchange(stat[nest],0));
      fprintf(out,"\n<H4>");
      fprintf(out,sectionnumbers(section)); }

    else if (eq(t->t,"\\LHendsubsection")) {
      fprintf(out,"</H4>\n<DIV>\n"); }

    else if (eq(t->t,"\\LHsubsubsection")) {
      section.s[3]++; section.level=SUBSUBSECTION;
      //      size=SIZE*4/3;
      size=SIZE;
      reopenhtml(htmlname(section),section);
      fprintf(out,fontchange(stat[nest],1));
      //      stat[nest].size=1; stat[nest].bold=0;
      stat[nest].size=0; stat[nest].bold=0; fontcolor=""; // void->css
      fprintf(out,"\n<H6>");
      fprintf(out,fontchange(stat[nest],0)); }

    else if (eq(t->t,"\\LHendsubsubsection")) {
      fprintf(out,"</H6>\n<DIV>"); }

    else if (eq(t->t,"\\LHmenu")) {
      out=outaux;
      size=SIZE;
      fprintf(out,"\n");
      fprintf(out,sectionlevel("ITEM",section)); }

    else if (eq(t->t,"\\endLHmenu")) {
      size=SIZE;
      fprintf(out,"-->\n"); out=outhtml; }

    else if (eq(t->t,"\\LHaudio")) {
      int it,nt=atoi(t->next->t);
      char *c;
      char name[128],*n,ext[8],*fn;
     
      t=t->next;
      do {
        t=t->next;
        c=t->t; }
      while (!isalnum(*c));

      n=name;
      while (isalnum(*c)||strchr("-_",*c)) {
        *n++=*c;
        t=t->next;
        c=t->t; }
      *n=0;
      
      do {
        t=t->next;
        c=t->t; }
      while (!isalnum(*c));

      n=ext;
      while (isalnum(*c)) {
        *n++=*c;
        t=t->next;
        c=t->t; }
      *n=0;
      
      fprintf(out,"<Div id=\"fl\">\n");
      loopto (it,1,nt) {
        fn=string("tabule/%s.%d.%s",name,it,ext);

        if (!fileexists(fn)) fprintf(stderr,"ERROR %s cannot be opened\n",fn);
        fprintf(out,"<img src=\"%s\" alt=\"%s\">\n",
                                fn,        fn);

        fn=string("audio/%s.%d.mp3",name,it);
        /* WARNING: sound-config.xml has baseHref="audio/"; cf. fn+6 below */
        if (!fileexists(fn)) fprintf(stderr,"ERROR %s cannot be opened\n",fn);
                
        if (soundconfig)
          fprintf(soundconfig,"<sound id=\"%s.%d\" href=\"%s\"/>\n",
                       shortbasename(section),it,         fn+6); }
                                
      fn=string("tabule/%s.%d.%s",name,nt+1,ext);
      if (fileexists(fn))
        fprintf(stderr,"WARNING %s exists but not referenced\n\
        (bad first arg to \\audio{}?)\n",fn);

      fn=string("audio/%s.%d.mp3",name,nt+1);
      if (fileexists(fn))
        fprintf(stderr,"WARNING %s exists but not referenced\n\
        (bad first arg to \\audio{}?)\n",fn);
      
      fprintf(out,"</Div><!-- id=\"fl\"-->\n"); 

      c=shortbasename(section);	
      fprintf(out,"<Div id=\"fr\">\n\
<ul id=\"vpr2\">\n\
  <li class=\"aud\"><a href=\"javascript:nosound('%s');\" id=\"prepzv\" title=\"Vypnout/zapnout zvuk (stiskni 2x pro opakovn)\"></a>\n\
  <li class=\"aud\" id=\"nasl\"><a href=\"javascript:audio(1,'%s');\" title=\"Nsledujc krok (t klvesa n nebo ipka vpravo)\"></a>\n\
  <li class=\"aud\" id=\"predch\"><a href=\"javascript:audio(0,'%s');\" title=\"Pedchoz krok (t klvesa p nebo ipka vlevo)\"></a>\n\
</ul>\n",c,c,c);
      }
      
    else if (eq(t->t,"\\BEGINTEXT"))
      fprintf(out,"<div>");
            
    else if (eq(t->t,"\\NEXTTEXT"))
      fprintf(out,"</div><div>");
            
    else if (eq(t->t,"\\ENDTEXT"))
      fprintf(out,"</div>\n\
</Div><!--id=\"fr\"-->\n\
<div class=\"cl\"></div>\n");
    
    else if (eq(t->t,"\\endLHaudio"))
      fprintf(out,"<!--\\audio-->\n");
        
    else if (eq(t->t,"\\scale")) {
      t=t->next;
      if (eq(t->t,"{")) {
	char nr[128];
	nr[0]=0;
	t=t->next;
	while (!eq(t->t,"}")) {
	  strcat(nr,t->t); 
	  t=t->next; }
	fprintf(stderr,"%s\n",nr);
	scale=atof(nr); }
      else tokerror(t,"{ expected after \\scale");
      if (scale<=0) tokerror(t,"negative or zero scale"); }

    /*** mathematics ***/
    else if (display || simplemath) {
      FILE *tex=NULL;
      tok_t *e,*tt;
      macro_t *m;
      uchar imgname[TOKENLEN];
      int crc,nchar,lchap,leqno;
      struct mathtex_s *mt,*mtx=NULL;

      if (envmath) e=findend(t);
      else if (epsfig) e=findright(t->next,"{")->next;
      else e=findright(t,t->t)->next;

      if (simplemath && !isgraphmath) {
        /* trying to expand simple mathematics into HTML */
        uchar line[16000];
        int nest=0,last=0,mbox=0,math=1,boldsymbol=0;
        stat_t stat[MAXNEST];

        memset(stat,0,sizeof(stat));
        stat[nest].font=IT;

        strcpy(line,fontchange(stat[nest],0));

        for (tt=t->next ;tt!=e; tt=tt->next) {
/*.....          fprintf(stderr,"%s\n",tt->t);*/

          if (tt->next==e) /* ending $ or \) ignored */ ;
          else if (eq(tt->t,"\\,"))
            strcat(line,SMALLSPACE);
          else if (eq(tt->t,"\\!")) 
            /* not supported and ignored */;
          else if (eq(tt->t,"\\;"))
            strcat(line,MIDSPACE);
          else if (eq(tt->t,"\\quad"))
            strcat(line,"&nbsp;&nbsp;&nbsp;");
          else if (eq(tt->t,"\\hspace")) {
            strcat(line,space(1,tt));
            tt=after->prev; }
          else if (eq(tt->t,"\\qquad"))
            strcat(line,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
          else if (isgreek(tt->t))
            strcat(line,greeksymbol);
          else if (eq(tt->t,"-"))
	    strcat(line,"&minus;");
          else if (eq(tt->t,"\\cdot"))
	    //	    strcat(line,"&sdot;"); // M$IE 6 does not know!
            strcat(line,"<sup> <b>.</b>&nbsp;</sup>");
	    //            strcat(line,"<font size=-1> <b>.</b>&nbsp;</font>");
          else if (eq(tt->t,"\\times"))
            strcat(line,"");
          else if (eq(tt->t," ") || eq(tt->t,"\n")) {
            if (!math) strcat(line," "); }
          else if (eq(tt->t,"$")) {
            /* math mode inside mbox */
            if (math && stat[nest].font==IT) {
              strcat(line,"</i>");
              stat[nest].font=RM; }
            if (!math && stat[nest].font!=IT) {
              strcat(line,"<i>");
              stat[nest].font=IT; } 
            math=!math; }
          else if (eq(tt->t,"~") || eq(tt->t,"\\ ") || eq(tt->t,"\\\n") || eq(tt->t,"\\\\"))
            /* \\ should not be here, but people often use it, so put at least space here... */
            strcat(line,"&nbsp;");
          else if (eq(tt->t,"\\circ")) {
            strcat(line,fontchange(stat[nest],1));
            strcat(line,"<sup><tt>o</tt></sup>");
            strcat(line,fontchange(stat[nest],0)); }
          else if (eq(tt->t,"{")) {
            strcat(line,fontchange(stat[nest],1));
            if (++nest>=MAXNEST)
              tokerror(tt,"simple math nesting exceeds MAXNEST");
            stat[nest]=stat[nest-1];
            if (last) {
              if (stat[nest].nie>=8)
                Error("_sub/^superscript nesting exceeded");
              stat[nest].ie[stat[nest].nie++]=last; }
            if (mbox) {
              stat[nest].font=RM;
              stat[nest].bold=stat[nest].underline=0; }
            if (boldsymbol) {
              stat[nest].font=IT;
              stat[nest].bold=1;
              stat[nest].underline=0; }
            mbox=boldsymbol=last=0;
            strcat(line,fontchange(stat[nest],0)); }
          else if (eq(tt->t,"}")) {
            strcat(line,fontchange(stat[nest],1));
            if (--nest<0) tokerror(tt,"math nesting");
            strcat(line,fontchange(stat[nest],0)); }
          else if (eq(tt->t,"_")) {
            if (strcmp(tt->next->t,"{")) tokerror(tt,"{ expected");
            last=SUB; }
          /* special hook for ^{\circ} ( don't use ^o !!! ) */
          else if (eq(tt->t,"^") && eq(tt->next->t,"{") 
                   && eq(tt->next->next->t,"\\circ") 
                   && eq(tt->next->next->next->t,"}") ) {
            strcat(line,"");
            tt=tt->next->next->next; }
          else if (eq(tt->t,"^")) {
            if (strcmp(tt->next->t,"{")) tokerror(tt,"{ expected");
            last=SUP; }
          else if (eq(tt->t,"\\mbox")) {
            if (strcmp(tt->next->t,"{"))
              tokerror(tt,"implementation limitation: argument of \\mbox must be in {}");
            mbox=1; math=0; }
          else if (eq(tt->t,"\\boldsymbol")) {
            if (strcmp(tt->next->t,"{"))
              tokerror(tt,"implementation limitation: argument of \\boldsymbol must be in {}");
            boldsymbol=1; math=1; }
          else if (eq(tt->t,"\\bf")) {
            /* bf sets also rm !!! - not general, but OK with simple math */
            /* use \boldsymbol{} for bold math italic */
            if (stat[nest].font==IT) {
              strcat(line,"</i>"); stat[nest].font=RM; }
            if (!(stat[nest].bold)) strcat(line,"<b>"); 
            stat[nest].bold=1; }
          else if (eq(tt->t,"\\LHul")) {
            if (!(stat[nest].underline)) strcat(line,"<u>"); 
            stat[nest].underline=1; }
          else if (eq(tt->t,"\\sf")) {
            if (stat[nest].font==IT) strcat(line,"</i>"); 
            else if (stat[nest].font) {
              fprintf(stderr,"simple math implementation limitation: font->sf\n");
              goto cannot; }
            stat[nest].font=SF;
            strcat(line,"<font face=\"helvetica,arial\">"); }
          else if (eq(tt->t,"\\tt")) {
            if (stat[nest].font==IT) strcat(line,"</i>"); 
            else if (stat[nest].font) {
              fprintf(stderr,"simple math implementation limitation: font->tt\n");
              goto cannot; }
            stat[nest].font=TT;
            strcat(line,"<font face=\"helvetica,arial\">"); }
          else if (eq(tt->t,"\\sin")
                || eq(tt->t,"\\cos")
                || eq(tt->t,"\\tan")
                || eq(tt->t,"\\log")
                || eq(tt->t,"\\ln")
                || eq(tt->t,"\\exp")
                || eq(tt->t,"\\max")
                || eq(tt->t,"\\min")) {
            if (stat[nest].font==IT)
              sprintf(strend(line),"</i>%s<i>" MIDSPACE,tt->t+1);
            else
              sprintf(strend(line)," %s" MIDSPACE,tt->t+1);
            }
          else if (eq(tt->t,"\\rm") || eq(tt->t,"\\textstyle") ) {
            if (stat[nest].font==IT) {
              strcat(line,"</i>"); stat[nest].font=RM; } }
          else if (isalpha(tt->t[0]) || tt->t[0]>=128 || eq(tt->t," "))
            strcat(line,tt->t);
          else if (isdigit(tt->t[0]) || strchr("+-/.,!\'[]()*|",tt->t[0])) {
            if (stat[nest].font==IT) sprintf(strend(line),"</i>%s<i>",tt->t);
            else strcat(line,tt->t); }
          else if (eq(tt->t,"="))
            strcat(line,EQSEP"="EQSEP);
          else if (eq(tt->t,"<"))
            strcat(line,EQSEP"&lt;"EQSEP);
          else if (eq(tt->t,">"))
            strcat(line,EQSEP"&gt;"EQSEP);
          else if (eq(tt->t,"\\dots") || eq(tt->t,"\\ldots"))
            strcat(line,DOTS);
          else goto cannot;
        }
        strcat(line,"</i>");

#ifdef DEBUG
        fprintf(stderr,"[ %s ]\n",line);
#endif
        { /* cleaning extra <X>..</X> pairs */
          uchar *c,*e;

          again:
            for (c=line; *c; )
              if (*c=='<') {
                if (*++c=='/') c++;
                
                for (e=c; *e; e++) {
                  while (*e!='>') e++;
                  e++;
                  if (*e!='<') break;
                  if (*++e=='/') e++;
                  if (*e=='s' || *e=='f') break; /* font,sub/sup cannot skip */
                  if (eqtag(c,e)) {
                    if (c[-1]==e[-1]) Error(line);
                    removetag(e);
                    removetag(c); 
                    goto again;
                  } } }
              else c++;
          }

#ifdef DEBUG
        fprintf(stderr,"[ %s ]\n",line);
#endif

        fprintf(out,"%s",line);
        t=e;
        goto endmath; }

     cannot:
      if (!isdebug && !isrerun) tex=fopen("aux.tex","wt");

      if (tex) {
        double x,y=(double)(simplemath?2:25);

/*.....	fprintf(stderr,"TeX: indent=%d\n",indent);*/
	if (display) x=(atoi(LINEWIDTH)-INDENT*3/2-indent*INDENT)/(double)DPI*2.54;
	/* since in BLOCKQUOTE, -INDENT*3/2 should be INDENT/2 from right */
	else x=12;
	if (x<5) x=5;
	if (x>19) tokerror(t,string("TeX->GIF conversion: %fcm is too wide page",x));

	fprintf(tex,"\\documentclass[12pt,fleqn]{book}\n\
\\usepackage{czech,epsfig,amsbsy}\n\
\\oddsidemargin -1cm\n\
\\textwidth=%.2fcm\n\
\\textheight=%.2fcm\n\
\\let\\d\\undefined\n\
\\pagestyle{empty}\n\
\\mathcode`,=\"002C\n\
\\renewcommand\\mathindent{0pt}\n\
\\newlength{\\keylen}\n\
\\let\\LHul=\\underline\n\
\\begin{document}\n",x,y); }

      looplist (m,macro0) if (!m->LH) {
        if (tex) {
          fprintf(tex,"\\newcommand{%s}[%d]{",m->name,m->n);
          looplist (tt,m->body)
            fprintf(tex,"%s%s",tt->t,afterspace(tt));
          fprintf(tex,"}\n"); } }

      //      fprintf(tex,"\\input ../unitchemr.tex\n");

/*.....      fprintf(tex,"\\setcounter{equation}{%d}\n",0);*/

      lchap=leqno=0;
      if (numbermath) {
        /* only equations with \label are numbered */
        int eqnarraynr=1,nest=0;
        for (tt=t->next ;tt!=e; tt=tt->next) {
          if (eq(tt->t,"\\begin")) nest++;
          if (eq(tt->t,"\\end")) nest--;
          if (eq(tt->t,"\\\\") && nest==0) {
            if (!eqnarray) tokerror(tt,"\\\\ in equation");
            eqnarraynr++; }
          if (eq(tt->t,"\\label")) {
            xref_t *x=findxref(tt->next->next->t);

            if (tex) {
              char *d=strchr(x->latexlabel,'.');

              lchap=atoi(x->latexlabel);
              if (!d) Error("internal: . expected in latexlabel");
              leqno=atoi(d+1)-eqnarraynr;
              fprintf(tex,"\\setcounter{chapter}{%d}\\setcounter{equation}{%d}\n",lchap,leqno); }

            fprintf(out,"<a name=\"%s\"></a>",x->name);
            goto number; } }

        /* no label ==> eq. not numbered */
        if (eq(t->next->next->t,"equation")) t->next->next->t=e->prev->prev->t="displaymath";
        if (eq(t->next->next->t,"eqnarray")) t->next->next->t=e->prev->prev->t="eqnarray*";

        number:; }

      finalres=(int)(size*scale+0.5); /* scaled size by \\scale{} */

      if (simplemath && tex) fprintf(tex,"\\framebox{\\rule[%fex]{0pt}{%fex}",belowex,aboveex-belowex);
      CRC_begin(); nchar=0;
      for ( ;t!=e; t=t->next) {
        uchar *c;

        nchar+=10000;
        for (c=t->t; *c; c++) { nchar++; CRC_next(*c); }
        if (tex) fprintf(tex,"%s%s",t->t,afterspace(t)); }
      crc=CRC();

      /* trying to find whether already the same math expression */
      looplist (mt,mathtex0)
        if (mt->crc==crc && mt->nchar==nchar && mt->size==finalres 
            && mt->lchap==lchap && mt->leqno==leqno) {
          mt->count++;
          if (tex) fclose(tex); /* discarded */
          goto mathfound; }

      /* trying to find whether available from previous run */
      looplist (mtx,matx0)
        if (mtx->crc==crc && mtx->nchar==nchar && mtx->size==finalres
            && mtx->lchap==lchap && mtx->leqno==leqno) break;

      allocone(mt);
      mt->next=mathtex0;
      mathtex0=mt;
      mt->crc=crc;
      mt->nchar=nchar;
      mt->size=finalres;
      mt->lchap=lchap;
      mt->leqno=leqno;
      mt->count=1;

      if (display&&simplemath) Error("internal");

      if (mtx) {
        /* reusing GIFs from previous run */
        char fn[LINELEN];
        char html[LINELEN];
        char *nr=strstr(mtx->html,"math/"),*x,type;

        if (!nr) Error("math.lst: no math/");
	x=strstr(mtx->html,"class=\"pos");
	bottom=x?atoi(x+10):0;
        strcpy(fn,nr);
        x=strchr(fn,'\"');
        if (!x) Error("math.lst: no \"");
        *x=0;
        type=fn[5];
        fn[3]='x'; /* math/ -> matx/ */
        nr+=6; *nr=0;
        sprintf(imgname,"math/%c%d.gif",type,nimg);
        fprintf(stderr,"moving \"%s\" -> \"%s\" ",fn,imgname);
        if (rename(fn,imgname)) Error("cannot move (erase both math.lst)");
        sprintf(html,"%s%d.gif\"%s",mtx->html,nimg,x+1);
        mt->html=mystrdup(html);
        if (tex) fclose(tex);
        tex=NULL; }
      else if (tex) {
        /* make GIFs from unresolved TeX */
        char s[LINELEN];

	if (finalres<10 || finalres>1200) tokerror(t,"wrong finalres");

        strcpy(s,string("cslatex aux; dvips -o aux.ps aux; gs -sDEVICE=%s -r%d -q -dSAFER -dNOPAUSE -sOutputFile=aux.ppm -I. -- aux.ps; l2htestr aux.ppm; ",ppmcolor?"ppmraw":"pgmraw",finalres));
	scale=1;

        if (simplemath) {       /* file    rt o% e% mod y */
	  /* NOTE: string MUST be in 1 line because written as comment to TeX */
          strcat(s,string("l2hcrop aux.ppm %d %d ;\
 ppmscale crop.ppm - 1/%d 1/%d "PPMSCALE" | %s > saux.gif ",
                          SCALE,SCALE, SCALE,SCALE, ppmtogif));
          if (simplemath) fprintf(tex,"}"); }
        else 
          strcat(s,string("ppmscale aux.ppm - 1/%d 1/%d "PPMSCALE" | pnmcrop | %s > daux.gif", 
                                              SCALE,SCALE,             ppmtogif));
          
        fprintf(tex,"\n%% %s\n\\end{document}\n",s);
        fclose(tex);

        fprintf(stderr,"sh %s\nretcode=%d\n>>>>> IMAGE %d FINISHED\n",s,system(s),nimg);

        {
          /* test margin + print progress indicator */
          FILE *right=fopen("l2htestr.aux","rt");
          tok_t *ttt=t->next;
          int i=0,n=0;

	  if (!right) Error("cannot open l2htestr.aux");

          fscanf(right,"%d%d",&n,&i);
          fclose(right);
          testr[0]=0;
          if (i>n-50) sprintf(testr," WATCH RIGHT MARGIN: %d of %d ",i,n); 

          n=i=0;
          looplist (ttt,head) { n++; if (t==ttt) i=n; };

          n=(79*i+n/2)/n;
          loop (i,0,79) { char c=i<n?'$':'-'; fputc(c,stderr); }
          fputc('\n',stderr); }

	/* pos of img */
        if (simplemath) {
          FILE *ct=fopen("l2hcrop.aux","rt");

          if (!ct) Error("l2hcrop.aux");
          fgets(ctline,LINELEN,ct);
          if (strend(ctline)[-1]=='\n') strend(ctline)[-1]='\n';
          bottom=atoi(ctline); /* class="pos#" */
	  if (bottom<0) bottom=0;
	  if (bottom>9) bottom=9;
          fclose(ct); }

        /* math/d#.gif : displaymath style 
           math/s#.gif : simple math, placed according to pos# style */
        if (simplemath) {
          sprintf(imgname,"math/s%d.gif",nimg);
          if (rename("saux.gif",imgname)) Error(imgname); }
        else {
          /* displaystyle */
          sprintf(imgname,"math/d%d.gif",nimg); 
          rename("daux.gif",imgname); } }
      else if (isdebug)
        sprintf(imgname,"math/%c%d.gif",simplemath?'$':'d',nimg);
      else {
        /* RERUN (!tex): finding imgname */
	Error("RERUN no longer supported"); }

      nimg++;

      if (isdebug)
        mt->html=mystrdup(string(" <font color=red>[%s]</font> ",imgname));
      else
        mt->html=mystrdup(imginfo(imgname,display?BLOCKQUOTE:POS,bottom,display?"displaymath":"$..$"));

     mathfound:
      fprintf(out,"%s",mt->html);
     endmath:
      t=t->prev; /* ??? */
      } /* math via gif */

    else if (eq(t->t,"\\,"))
      fprintf(out,SMALLSPACE);

    else if (eq(t->t,"\\obrazek") || eq(t->t,"\\obrazekvpravo") || eq(t->t,"\\obrazekvlevo")) {
      char *figname=string("fig/%s",getgroup(t->next));

      fprintf(out,imginfo(figname,eq(t->t,"\\obrazekvpravo")?RIGHT:eq(t->t,"\\obrazekvlevo")?LEFT:0,0,figname));
      t=after->prev; }      

    else if (eq(t->t,"\\html")) {
      fprintf(out,getgroup(t->next)); 
      t=after->prev; }      

    else if (isbegin(t,"table")) { fprintf(out,"<!--begin{table}: float ignored-->"); t=t->next->next->next; }
    else if (isend(t,"table")) { fprintf(out,"<!--end{table}: float ignored-->"); t=t->next->next->next; }

    else if (isbegin(t,"center")) { fprintf(out,"<center>"); t=t->next->next->next; indent++; }
    else if (isend(t,"center")) { fprintf(out,"</center>"); t=t->next->next->next; indent--; }

#if 0
    else if (isbegin(t,"quote")) { fprintf(out,"<blockquote>"); t=t->next->next->next; indent++; }
    else if (isend(t,"quote")) { fprintf(out,"</blockquote>"); t=t->next->next->next; indent--; }
#else
    /* NOTE: environment quote cannot be translated by <blockquote> because
       some browsers (M$IE5.0) make a right margin which interacts
       with equations */
    else if (isbegin(t,"quote")) {
      fprintf(out,string("<p><table border=0 cellspacing=0 cellpadding=0 width=%d><tr><td width=%d>&nbsp;</td><td width=%d>",atoi(LINEWIDTH)-INDENT*indent,INDENT,atoi(LINEWIDTH)-INDENT*(indent+1)));
      t=t->next->next->next; indent++; }
    else if (isend(t,"quote")) {
      fprintf(out,"</td></tr></table><p>");
      t=t->next->next->next; indent--; }
#endif

#if 0
    // itemize, enumerate, abc by means of html
    else if (isbegin(t,"itemize")) { fprintf(out,"<ul class=\"bullet\">"); t=t->next->next->next; indent++; }
    else if (isend(t,"itemize")) { fprintf(out,"</ul>"); t=t->next->next->next; indent--; }

    else if (isbegin(t,"enumerate")) { fprintf(out,"<ol class=\"decimal\">"); t=t->next->next->next; indent++; }
    else if (isend(t,"enumerate")) { fprintf(out,"</ol>"); t=t->next->next->next; indent--; }

    else if (isbegin(t,"abc")) { fprintf(out,"<ol class=\"abc\">"); t=t->next->next->next; indent++; }
    else if (isend(t,"abc")) { fprintf(out,"</ol>"); t=t->next->next->next; indent--; }

    else if (eq(t->t,"\\item")) fprintf(out,"<li>");
#else
    // only one level
    else if (isbegin(t,"itemize")) { fprintf(out,"<!--itemize-->"); t=t->next->next->next; indent++; list.type='.'; list.n=0; }
    else if (isend(t,"itemize"))   { fprintf(out,"</div><!--/itemize-->\n"); t=t->next->next->next; indent--; }

    else if (isbegin(t,"enumerate")) { fprintf(out,"<!--enumerate-->"); t=t->next->next->next; indent++; list.type='1'; list.n=0; }
    else if (isend(t,"enumerate"))   { fprintf(out,"</div><!--/enumerate-->\n"); t=t->next->next->next; indent--; }

    else if (isbegin(t,"abc")) { fprintf(out,"<!--itemize-->"); t=t->next->next->next; indent++; list.type='a'; list.n=0; }
    else if (isend(t,"abc"))    { fprintf(out,"</div><!--/itemize-->\n"); t=t->next->next->next; indent--; }

    else if (eq(t->t,"\\item")) {
      if (list.n) fprintf(out,"</div>\n");
      fprintf(out,"<div class=\"item\">");
      list.n++;
      switch (list.type) {
        case '.': fprintf(out,"&nbsp;<font size=\"+1\">&bull;</font>&nbsp; "); break;
        case '1': fprintf(out,"%d.&nbsp; ",list.n); break;
        case 'a': fprintf(out,"%c)&nbsp; ",list.n+'`'); break;
        default: tokerror(t,"list"); } }
#endif

    else if (isbegin(t,"htmltable") || isbegin(t,"longtable")) {
      /* WARNING: nesting of tabular/htmltable probably wrong */
      int n;

      if (htmltable>=4) tokerror(t,"htmltable/longtable nesting>4");
      t=t->next->next;
      while (strcmp(t->t,"{")) t=t->next;
      t=t->next;
      tab[htmltable].n=0;
      tab[htmltable].border=0;
      loop (n,0,32) tab[htmltable].bg[n]=0;
      while (strcmp(t->t,"}")) {
        if (tab[htmltable].n>=32) tokerror(t,"htmltable/longtable columns>32");
        if (t->t[0]>' ') {
          if (eq(t->t,"c")) tab[htmltable].align[tab[htmltable].n++]="center";
          else if (eq(t->t,"l")) tab[htmltable].align[tab[htmltable].n++]="left";
          else if (eq(t->t,"r")) tab[htmltable].align[tab[htmltable].n++]="right";
          else if (eq(t->t,"|")) {
            tab[htmltable].border=1; /* at least one | causes border */
	    tab[htmltable].bg[tab[htmltable].n-1]++; 
	    tab[htmltable].bg[tab[htmltable].n]++; }
          else tokerror(t,"htmltable/longtable unsupported align specifier"); }
        t=t->next; }
      
      tab[htmltable].i=0;
      htmltable++;
      fprintf(out,"<table cellspacing=2 cellpadding=2 border=%d><tr bgcolor=red>",tab[htmltable-1].border);
      emptycol=1; /* BUG: should be in tab[] because of nesting */
      htmltabtd(); }
    else if (isend(t,"htmltable") || isend(t,"longtable")) {
      htmltable--;
      titlerow=0;
      if (htmltable<0) tokerror(t,"unexpected end of htmltable/longtable");
      fprintf(out,"</td></tr></table>\n");
      t=t->next->next->next; }
    else if (eq(t->t,"\\titlerow")) titlerow++; /* color bg of a row */

    else if (eq(t->t,"\\par") || eq(t->t,"\n\n"))
      fprintf(out,"<p>");
/*.....      fprintf(out,"<br>&nbsp;<br>");*/
    else if (eq(t->t,"\"") || eq(t->t,"''") || eq(t->t,"``"))
      fprintf(out,"&quot;");
    else if (eq(t->t,"\\&"))
      fprintf(out,"&amp;");
    else if (specchar(t->t))
      fprintf(out,"%s",t->t+1);
    else if (eq(t->t,"\\mbox")) ;
    else if (eq(t->t,"~")) fprintf(out,"&nbsp;");
    else if (eq(t->t,"\\ ") || eq(t->t,"\\\n")) fprintf(out,eq(t->next->t,"\\ ")||eq(t->t,"\\\n") ? "&nbsp;" : " ");
    else if (eq(t->t,"\\dots") || eq(t->t,"\\ldots"))
      fprintf(out,DOTS);
/*.....    else if (eq(t->t,"--")) fprintf(out,"<font size=+1>-</font>");*/
    else if (eq(t->t,"--")) fprintf(out,"&ndash;");
    else if (eq(t->t,"---")) fprintf(out," &mdash; ");
    else if (eq(t->t,"\\quad")) fprintf(out,"&ensp;");
    else if (eq(t->t,"\\qquad")) fprintf(out,"&emsp;");
    else if (eq(t->t,"\\hspace")) {
      fprintf(out,space(1,t));
      t=after->prev; }
    else if (eq(t->t,"\\vspace")) {
      fprintf(out,space(0,t));
      t=after->prev; }
    else if (eq(t->t,"&")) {
      if (htmltable<=0) tokerror(t,"unexpected &");
      if (tab[htmltable-1].i>=32) tokerror(t,"too many &\'s in table row");
      if (emptycol) fprintf(out,"&nbsp;"); emptycol=1;
      fprintf(out,"</td>");
      htmltabtd(); }
    else if (eq(t->t,"\\\\") || eq(t->t,"\\linebreak")) {
      if (eq(t->t,"\\\\") && htmltable) {
	/* testing, whether to ignore the last \\ */
	tok_t *tt=t->next;
	
	titlerow=0;
	while (tt->t[0]<=' ' || eq(tt->t,"\\hline")) tt=tt->next;
	if (isend(tt,"htmltable") || isend(tt,"longtable")) /* nothing */;
	else {
	  tab[htmltable-1].i=0;
	  if (emptycol) fprintf(out,"&nbsp;"); emptycol=1;
	  fprintf(out,"</td></tr><tr>"); 
	  if (eq(t->next->t,"\\titlerow") || eq(t->next->next->t,"\\titlerow")) titlerow=1;
	  htmltabtd(); } }
      else fprintf(out,"<br>\n");
      if (eq(t->next->t,"[")) {
        t=t->next->next;
        if (!eq(t->t,"-")) fprintf(out,"<br>\n");
        while (!eq(t->t,"]")) t=t->next; } }
    else if (t->t[0]=='\\') {
      struct ignored_s *i;

      looplist (i,ignored0) if (eq(i->name,t->t)) goto ifound;
      allocone(i);
      i->name=t->t;
      i->next=ignored0; ignored0=i; 
    ifound:
      fprintf(stderr,"WARNING command %s ignored\n",t->t); }
    else if (eq(t->t,"{") || eq(t->t,"}"))
      /* any extra {} ignored */ ;

    else
      fprintf(out,"%s",t->t); }
  /* >>>>> */

  closehtml();
  fclose(outaux);
  fclose(openedimages);
  if (cropinfo) fclose(cropinfo);
  if (soundconfig) {
    fprintf(soundconfig,"</items>\n");
    fclose(soundconfig); }

  /* uklada informace o obrazcich */
  if (!isdebug && !isrerun) {
    FILE *mtf=fopen("math.lst","wt");
    struct mathtex_s *mt;

    fprintf(mtf,"# CRC  nchar size lchap leqno count CR html .. CR CR\n");

    looplist (mt,mathtex0)
      fprintf(mtf,"%8x %2d %3d %d %d %d\n%s\n\n",
              mt->crc,mt->nchar,mt->size,mt->lchap,mt->leqno,mt->count,mt->html);

    fclose(mtf); }

  {
    FILE *f=fopen("ignored.lst","wt");
    struct ignored_s *i;

    looplist (i,ignored0) fprintf(f,"%s\n",i->name); 
    fclose(f); }

  if (iszip) system(string("zip %szip %s*html",fileprefix,fileprefix));

  fprintf(stderr,"pass 1 finished\n");
  sleep(2);

  /* 2nd pass */
  system(string("l2h2 %d %s*html",shorts,fileprefix));

  return 0;
}
