/*  cc -o text -O2 text.c

  text file standardizer

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char uchar;

void Error(uchar *msg)
{
  fprintf(stderr,"ERROR %s\n",msg);
  exit(-1);
}

#define LLEN 10242

static uchar line[LLEN],outline[LLEN+80*2];

int tabl=8;
int padd=-1;
int r7=0;

uchar *myfgets(uchar *l,int nl,FILE *in)
{
  int i=0,c;

  l[i]=0;

  for (;;) {
    c=fgetc(in);
    if (c<0) {
      if (i) return l;
      else return NULL; }
    if (c==10 || c==13) {
      int cc=fgetc(in);
      
      if (cc<0) return l; /* EOF after CR or LF */
      
      if (c==10) {
	/* ignore any number of CR after LF */
	while (cc==13) cc=fgetc(in);
	if (cc>0) if (ungetc(cc,in)<0) Error("ungetc");
	return l; }
      
      if (c==13) {
	/* ignore any number of LF after CR */
	while (cc==10) cc=fgetc(in);
	if (cc>0) if (ungetc(cc,in)<0) Error("ungetc");
	return l; }
      }

  l[i++]=c;
  l[i]=0;
  if (i>=nl) Error("input line too long"); }
}

void dofile(char *fn)
{
  FILE *in,*out;
  uchar *ch;
  int filter=!strcmp(fn,"-");
  int lc;

  if (filter)
    in=stdin,out=out;
  else {
    in=fopen(fn,"rt");
    printf("opening %s\n",fn);
    if (!in) Error(fn);
    out=fopen("text.aux","wt");
    if (!out) Error("text.aux"); }

  while (myfgets(line,LLEN-1,in)) {
    
    /* remove the highest bit */
    if (r7) for (ch=line; *ch; ch++) *ch&=127;
    
    memset(outline,0,LLEN+80*2);
    strcpy(outline,line);

    /* expand tabs */
    if (tabl) {
      lc=0;
      for (ch=line; *ch; ch++) {
	if (*ch!='\t') outline[lc++]=*ch;
	else do outline[lc++]=' '; while (lc%tabl);
	if (lc>=LLEN+80) Error("output line too long"); } }
    
    /* remove padding spaces */
    if (padd) {
      lc=strlen(outline)-1;
      while (lc>=0 && outline[lc]==' ') outline[lc--]=0; }
    
    /* fill upto padd by spaces */
    if (padd>0) while (strlen(outline)<padd) strcat(outline," ");
    
    /* add [CR]LF */
    strcat(outline,"\n");
    
    fputs(outline,out); }
  
  fclose(out);

  if (!filter) {
    sprintf(line,"%s~",fn);
    if (rename(fn,line)) Error("make backup");
    if (rename("text.aux",fn)) Error("rename text.aux"); }
}

int main(int narg,char **arg)
{
  int i;

  if (narg<2) {
    fprintf(stderr,"Standardize text file.  Call by:\n\
  text {OPTION|FILE|-} [{OPTION|FILE|-} ...]\n\
Default: expand tabs (tab len=8), remove padding spaces\n\
Always: standardize EOL (inc. last line)\n\
OPTION:\n\
  -t# expand tabs, tab len=#\n\
  -t0 don\'t expand tabs\n\
  -p# adjust padding spaces to make all lines at least # char long\n\
  -p0 don\'t touch padding spaces\n\
  -r1 remove the highest bit\n\
FILE: backup FILE to FILE~ and standardize FILE\n\
-: stdin to stdout\n");
    exit(0); }

  for (i=1; i<narg; i++)
    if (arg[i][0]=='-')
      switch (arg[i][1]) {
      case 't':
        tabl=atoi(arg[i]+2);
	if (tabl>80 || tabl<0) Error("bad value of -t");
	break;
      case 'p':
	padd=atoi(arg[i]+2);
	if (padd>LLEN+80) Error("bad value of -p");
	break;
      case 'r':
	r7=atoi(arg[i]+2);
	break;
      case 0: 
        dofile("-");
        break;
      default: Error("bad option"); }
    else 
      dofile(arg[i]);

  return 0;
}
